@@ -139,7 +139,18 @@ class CachingRoleProvider(RoleProvider):
 
         # Cache the roles to which the environ is known to belong or not,
         cached_is_in |= member_roles
-        cached_not_in |= not_cached - member_roles
+
+        if provider_lookup is self.provider.member_subset:
+            # member_subset can be relied on to return all roles for a member,
+            # so we can safely infer negative roles
+            cached_not_in |= not_cached - member_roles
+        else:
+            # with ``any_matching`` we can only infer negative membership in
+            # the case that no roles matched (any_matching stops after the
+            # first hit, so we do not know if the user may have other roles
+            # that were not tested)
+            if len(member_roles) == 0:
+                cached_not_in |= not_cached
 
         return roles & cached_is_in
 
